#!/usr/bin/env python3
"""Test the converter module."""

# Standard imports
import unittest
import os
import sys
import time

# Try to create a working PYTHONPATH
EXEC_DIR = os.path.dirname(os.path.realpath(__file__))
ROOT_DIR = os.path.abspath(os.path.join(
    os.path.abspath(os.path.join(EXEC_DIR, os.pardir)), os.pardir))
if EXEC_DIR.endswith('/pattoo-shared/tests/test_pattoo_shared') is True:
    # We need to prepend the path in case PattooShared has been installed
    # elsewhere on the system using PIP. This could corrupt expected results
    sys.path.insert(0, ROOT_DIR)
else:
    print('''\
This script is not installed in the "pattoo-shared/tests/test_pattoo_shared" \
directory. Please fix.''')
    sys.exit(2)

# Pattoo imports
from pattoo_shared import converter
from pattoo_shared.configuration import Config
from pattoo_shared.variables import (
    DataPointMetadata, DataPoint, DeviceDataPoints, AgentPolledData)
from pattoo_shared.constants import (
    DATA_FLOAT, DATA_INT, DATA_COUNT64, DATA_COUNT, DATA_STRING, DATA_NONE,
    DATAPOINT_KEYS)
from tests.libraries.configuration import UnittestConfig


class TestBasicFunctions(unittest.TestCase):
    """Checks all functions and methods."""

    #########################################################################
    # General object setup
    #########################################################################

    config = Config()

    def test_cache_to_keypairs(self):
        """Testing method / function cache_to_keypairs."""
        self.maxDiff = None

        cache = {
            'pattoo_source': '1234',
            'pattoo_polling_interval': 30,
            'pattoo_datapoints': [
                {'pattoo_metadata': [
                    {'pattoo_agent_hostname': 'palisadoes'},
                    {'pattoo_agent_id': '1234'},
                    {'pattoo_agent_program': 'program_1'},
                    {'pattoo_device': 'device_1'},
                    {'gateway': 'palisadoes'}],
                 'pattoo_key': 30386,
                 'pattoo_data_type': 99,
                 'pattoo_value': 523.0,
                 'pattoo_timestamp': 1574011824387,
                 'pattoo_checksum': '123'}
            ]
        }

        # Test
        results = converter.cache_to_keypairs(cache)
        self.assertTrue(isinstance(results, list))
        for _, result in enumerate(results):
            self.assertEqual(result.pattoo_checksum, '''\
5c531afb2e4106cc78f99ff895ef99d2fa587f7272a91f82244f06caddf89176c299afc0ef4a44\
7d6c5042dff4690e271a3d09aabf97465faba20591f818ab27''')
            self.assertEqual(result.pattoo_timestamp, 1574011824387)
            self.assertEqual(result.pattoo_value, 523.0)
            self.assertEqual(result.pattoo_polling_interval, 30000)
            self.assertEqual(result.pattoo_data_type, 99)
            self.assertEqual(result.pattoo_key, 30386)
            _metadata = cache['pattoo_datapoints'][0]['pattoo_metadata']
            self.assertEqual(
                result.pattoo_metadata,
                [(_k_, _v_) for _dict in _metadata for _k_, _v_ in sorted(
                    _dict.items())])

        # Test for all types of data types
        for data_type in [DATA_FLOAT, DATA_INT, DATA_COUNT64, DATA_COUNT]:
            cache = {
                'pattoo_source': '1234',
                'pattoo_polling_interval': 30,
                'pattoo_datapoints': [
                    {'pattoo_metadata': [
                        {'pattoo_agent_hostname': 'palisadoes'},
                        {'pattoo_agent_id': '1234'},
                        {'pattoo_agent_program': 'program_1'},
                        {'pattoo_device': 'device_1'},
                        {'gateway': 'palisadoes'}],
                     'pattoo_key': 30386,
                     'pattoo_data_type': data_type,
                     'pattoo_value': 523.0,
                     'pattoo_timestamp': 1574011824387,
                     'pattoo_checksum': '123'}
                ]
            }

            # Test
            results = converter.cache_to_keypairs(cache)
            self.assertTrue(isinstance(results, list))
            for _, result in enumerate(results):
                self.assertEqual(result.pattoo_timestamp, 1574011824387)
                self.assertEqual(result.pattoo_value, 523.0)
                self.assertEqual(result.pattoo_polling_interval, 30000)
                self.assertEqual(result.pattoo_data_type, data_type)
                self.assertEqual(result.pattoo_key, 30386)
                _metadata = cache['pattoo_datapoints'][0]['pattoo_metadata']
                self.assertEqual(
                    result.pattoo_metadata,
                    [(_k_, _v_) for _dict in _metadata for _k_, _v_ in sorted(
                        _dict.items())])

    def test_agentdata_to_datapoints(self):
        """Testing method / function agentdata_to_datapoints."""
        # Setup AgentPolledData
        agent_program = 'panda_bear'
        apd = AgentPolledData(agent_program, self.config)

        # Initialize DeviceDataPoints
        device = 'teddy_bear'
        ddv = DeviceDataPoints(device)

        # Setup DataPoint
        value = 457
        key = 'gummy_bear'
        data_type = DATA_INT
        variable = DataPoint(key, value, data_type=data_type)

        # Add data to DeviceDataPoints
        ddv.add(variable)

        # Test DeviceGateway to AgentPolledData
        apd.add(ddv)

        # Test contents
        expected_metadata = {
            'pattoo_agent_id': apd.agent_id,
            'pattoo_agent_program': agent_program,
            'pattoo_agent_hostname': apd.agent_hostname,
            'pattoo_agent_polled_device': device
        }
        result = converter.agentdata_to_datapoints(apd)

        self.assertEqual(len(result), 1)
        item = result[0]
        self.assertTrue(isinstance(item, DataPoint))
        self.assertEqual(item.value, value)
        self.assertEqual(item.data_type, DATA_INT)
        self.assertEqual(item.key, key)
        self.assertTrue(isinstance(item.metadata, dict))
        self.assertEqual(len(item.metadata), len(expected_metadata))
        for key, value in item.metadata.items():
            self.assertTrue(isinstance(value, str))
            self.assertTrue(isinstance(key, str))
            self.assertEqual(value, str(expected_metadata[key]))

    def test_datapoints_to_dicts(self):
        """Testing method / function datapoints_to_dicts."""
        # Initialize key variables
        datapoints = []
        now = time.time()

        # Create DataPoints
        for value in range(0, 5):
            metadata = []
            for meta in range(0, 22, 7):
                metadata.append(DataPointMetadata(int(meta), str(meta * 2)))

            # Create the datapoint
            datapoint = DataPoint(
                'label_{}'.format(value), value,
                data_type=('type_{}'.format(value))
            )
            # Add metadata
            for meta in metadata:
                datapoint.add(meta)

            # Add metadata that should be ignored.
            for key in DATAPOINT_KEYS:
                metadata.append(DataPointMetadata(key, '_{}_'.format(key)))

            # Add the datapoint to the list
            datapoints.append(datapoint)

        # Start testing
        result = converter.datapoints_to_dicts(datapoints)
        self.assertTrue(isinstance(result, list))
        for value, item in enumerate(result):
            self.assertEqual(item['pattoo_data_type'], 'type_{}'.format(value))
            self.assertEqual(item['pattoo_key'], 'label_{}'.format(value))
            self.assertTrue(item['pattoo_timestamp'] >= now)
            self.assertTrue(item['pattoo_timestamp'] <= now + 1000)
            self.assertTrue(isinstance(item['pattoo_checksum'], str))
            self.assertTrue(isinstance(item['pattoo_metadata'], list))
            self.assertTrue(len(item['pattoo_metadata'], 4))
            for meta, metadata in enumerate(item['pattoo_metadata']):
                self.assertTrue(metadata, dict)
                for m_key, m_value in metadata.items():
                    self.assertEqual(2 * int(m_key), int(m_value))
                    self.assertEqual(int(m_key) % 7, 0)

    def test_agentdata_to_post(self):
        """Testing method / function agentdata_to_post."""
        # Setup AgentPolledData
        agent_program = 'panda_bear'
        polling_interval = self.config.polling_interval()
        apd = AgentPolledData(agent_program, self.config)

        # Initialize DeviceDataPoints
        device = 'teddy_bear'
        ddv = DeviceDataPoints(device)

        # Setup DataPoint
        value = 457
        key = 'gummy_bear'
        data_type = DATA_INT
        variable = DataPoint(key, value, data_type=data_type)

        # Add data to DeviceDataPoints
        ddv.add(variable)

        # Test DeviceGateway to AgentPolledData
        apd.add(ddv)
        result = converter.agentdata_to_post(apd)
        self.assertEqual(result.pattoo_source, apd.agent_id)
        self.assertEqual(
            result.pattoo_polling_interval, polling_interval * 1000)
        self.assertTrue(isinstance(result.pattoo_datapoints, list))

        # We have a dict to evaluate
        datapoint = result.pattoo_datapoints[0]
        self.assertTrue(isinstance(datapoint, dict))
        self.assertTrue(
            datapoint['pattoo_checksum'],
            'a488c71cafa214ee81f670eb0f935dc809374daee0664fe815f28ea628c3c8b3')
        self.assertTrue(datapoint['pattoo_data_type'], DATA_INT)
        self.assertTrue(datapoint['pattoo_key'], key)
        self.assertTrue(datapoint['pattoo_value'], value)

        expected_metadata = {
            'pattoo_agent_id': apd.agent_id,
            'pattoo_agent_program': apd.agent_program,
            'pattoo_agent_hostname': apd.agent_hostname,
            'pattoo_agent_polled_device': device
        }
        for item in datapoint['pattoo_metadata']:
            for key, value in item.items():
                self.assertTrue(isinstance(value, str))
                self.assertTrue(isinstance(key, str))
                self.assertEqual(value, str(expected_metadata[key]))

    def test_datapoints_to_post(self):
        """Testing method / function datapoints_to_post."""
        # Initialize key variables
        key = '_key'
        value = '_value'
        datapoints = [DataPoint(key, value)]
        source = '1234'
        polling_interval = self.config.polling_interval()
        result = converter.datapoints_to_post(
            source, polling_interval, datapoints)

        # Test
        self.assertEqual(result.pattoo_polling_interval, polling_interval)
        self.assertEqual(result.pattoo_source, source)
        self.assertEqual(result.pattoo_datapoints, datapoints)
        self.assertEqual(result.pattoo_datapoints[0].key, key)
        self.assertEqual(result.pattoo_datapoints[0].value, value)

    def test_posting_data_points(self):
        """Testing method / function posting_data_points."""
        # Initialize key variables
        key = '_key'
        value = '_value'
        datapoints = [DataPoint(key, value)]
        source = '1234'
        polling_interval = self.config.polling_interval()
        pdp = converter.datapoints_to_post(
            source, polling_interval, datapoints)
        result = converter.posting_data_points(pdp)

        # Test
        self.assertEqual(result['pattoo_polling_interval'], polling_interval)
        self.assertEqual(result['pattoo_source'], source)
        self.assertEqual(result['pattoo_datapoints'], datapoints)
        self.assertEqual(result['pattoo_datapoints'][0].key, key)
        self.assertEqual(result['pattoo_datapoints'][0].value, value)


if __name__ == '__main__':
    # Make sure the environment is OK to run unittests
    UnittestConfig().create()

    # Do the unit test
    unittest.main()



#!/usr/bin/env python3
"""Pattoo Data Converter."""

# Standard imports
import re

# Pattoo libraries
from .variables import (
    ConverterMetadata, DataPoint, AgentPolledData,
    PostingDataPoints)
from .constants import (
    DATA_FLOAT, DATA_INT, DATA_COUNT64, DATA_COUNT, DATA_STRING, DATA_NONE,
    MAX_KEYPAIR_LENGTH, PattooDBrecord, RESERVED_KEYS, DATAPOINT_KEYS,
    CACHE_KEYS)
from pattoo_shared import data
from pattoo_shared import log


def cache_to_keypairs(_data):
    """Convert agent cache data to AgentPolledData object.

    Args:
        items: Cache data
        source: Source of the cache data

    Returns:
        result: Validated cache data. [] if invalid.

    """
    # Initialize key variables
    result = []
    _log_message = 'Invalid cache data.'

    # Basic validation
    if isinstance(_data, dict) is False:
        log.log2warning(1032, _log_message)
        return []
    if len(_data) != len(CACHE_KEYS):
        log.log2warning(1033, _log_message)
        return []
    for key in _data.keys():
        if key not in CACHE_KEYS:
            log.log2warning(1034, _log_message)
            return []

    # Intialize variables needed later
    polling_interval = _data['pattoo_polling_interval']
    items = _data['pattoo_datapoints']
    source = _data['pattoo_source']

    # Verify we are getting a list
    if isinstance(items, list) is False:
        log.log2warning(1035, _log_message)
        return []

    # Verify contents of lists
    for item in items:
        # Initialize data
        valids = []
        keypairs = []

        # Verify dicts are in the list
        if isinstance(item, dict) is False:
            continue

        # Get all the key-pairs for the item
        keypairs = []
        for key, value in sorted(item.items()):
            # All the right keys
            valids.append(key in DATAPOINT_KEYS)
            valids.append(len(item) == len(DATAPOINT_KEYS))
            valids.append(isinstance(key, str))

            # Work on metadata
            if key == 'pattoo_metadata':
                # Must be a dict
                if isinstance(value, list) is False:
                    valids.append(False)
                    continue

                # Add metadata keypairs as a list of tuples
                for keypair_dict in value:
                    if isinstance(keypair_dict, dict) is False:
                        continue
                    keypairs.extend(_keypairs(keypair_dict, RESERVED_KEYS))

            # Work on the data_type
            if key == 'pattoo_data_type':
                valids.append(value in [
                    DATA_FLOAT, DATA_INT, DATA_COUNT64, DATA_COUNT,
                    DATA_STRING, DATA_NONE])

        # Append to result
        if False not in valids:
            # Add the datasource to the original checksum for better uniqueness
            """checksum = data.hashstring(
                '{}{}'.format(source, item['pattoo_checksum']), sha=512)"""
            checksum = _checksum(source, item)
            pattoo_db_variable = PattooDBrecord(
                pattoo_checksum=checksum,
                pattoo_key=item['pattoo_key'],
                pattoo_source=source,
                pattoo_polling_interval=polling_interval,
                pattoo_timestamp=item['pattoo_timestamp'],
                pattoo_data_type=item['pattoo_data_type'],
                pattoo_value=item['pattoo_value'],
                pattoo_metadata=keypairs
            )
            result.append(pattoo_db_variable)

    # Return
    return result


def agentdata_to_datapoints(agentdata):
    """Ingest data.

    Args:
        agentdata: AgentPolledData object

    Returns:
        rows: List of DataPoint objects

    """
    # Initialize key variables
    rows = []

    # Only process valid data
    if isinstance(agentdata, AgentPolledData) is True:
        # Return if invalid data
        if bool(agentdata.valid) is False:
            return []

        for ddv in agentdata.data:
            # Ignore bad data
            if ddv.valid is False:
                continue

            # Get data
            for _dv in ddv.data:
                # Assign values to DataPoints
                metadata = {
                    'pattoo_agent_id': agentdata.agent_id,
                    'pattoo_agent_program': agentdata.agent_program,
                    'pattoo_agent_hostname': agentdata.agent_hostname,
                    'pattoo_agent_polled_device': ddv.device
                }
                for key, value in sorted(metadata.items()):
                    _dv.add(ConverterMetadata(key, value))
                rows.append(_dv)

    # Return
    return rows


def datapoints_to_dicts(items):
    """Ingest data.

    Args:
        items: List of datapoints to convert

    Returns:
        result: List of datapoints converted to a list of dicts

    """
    # Initialize key variables
    datapoints = []
    result = []

    # Verify input data
    if isinstance(items, list) is False:
        items = [items]
    for item in items:
        if isinstance(item, DataPoint):
            datapoints.append(item)

    # Convert to a list of dicts
    for datapoint in datapoints:
        # Only convert valid data
        if datapoint.valid is True:
            data_dict = {
                'pattoo_metadata': [
                    {str(key): str(value)} for key, value in sorted(
                        datapoint.metadata.items())],
                'pattoo_key': datapoint.key,
                'pattoo_data_type': datapoint.data_type,
                'pattoo_value': datapoint.value,
                'pattoo_timestamp': datapoint.timestamp,
                'pattoo_checksum': datapoint.checksum
            }
            result.append(data_dict)

    return result


def agentdata_to_post(agentdata):
    """Create data to post to the pattoo API.

    Args:
        agentdata: AgentPolledData object

    Returns:
        result: Dict of data to post

    """
    # Initialize key Variables
    source = agentdata.agent_id
    polling_interval = agentdata.polling_interval
    _data = agentdata_to_datapoints(agentdata)
    _datapoints = datapoints_to_dicts(_data)
    result = datapoints_to_post(source, polling_interval, _datapoints)
    return result


def datapoints_to_post(source, polling_interval, datapoints):
    """Create data to post to the pattoo API.

    Args:
        source: Unique source ID string
        polling_interval: Interval over which the data is periodically polled
        datapoints: List of DataPoint objects

    Returns:
        result: Dict of data to post

    """
    result = PostingDataPoints(source, polling_interval, datapoints)
    return result


def posting_data_points(_data):
    """Create data to post to the pattoo API.

    Args:
        _data: PostingDataPoints object

    Returns:
        result: Dict of data to post

    """
    result = {
        'pattoo_source_timestamp': _data.pattoo_timestamp,
        'pattoo_source': _data.pattoo_source,
        'pattoo_polling_interval': _data.pattoo_polling_interval,
        'pattoo_datapoints': _data.pattoo_datapoints}
    return result


def _keypairs(_data, exclude_list):
    """Make key-pairs from metadata dict.

    Args:
        data: Metadata dict

    Returns:
        result: List of tuples of key-pair values

    """
    # Initialize key variables
    result = []

    # Loop around keys
    for _key, value in _data.items():
        # We want to make sure that we don't have
        # duplicate key-value pairs
        if _key in exclude_list:
            continue
        # Key-Value pairs must be strings
        if isinstance(_key, str) is False or isinstance(
                value, str) is False:
            continue

        # Standardize the keys
        splits = re.findall(r"[\w']+", _key)
        key = '_'.join(splits).lower()

        # Update the list
        result.append(
            (str(key)[:MAX_KEYPAIR_LENGTH], str(
                value)[:MAX_KEYPAIR_LENGTH])
        )

    return result


def _checksum(identifier, record):
    """Create a unique checksum for a DataPoint based on agent and device.

    Args:
        identifier: Agent ID that reported the datapoints
        record: PattooDBrecord converted to a Dict

    Returns:
        result: Checksum

    """
    # Initialize key variables
    device = None
    agent_id = None
    d_key = 'pattoo_agent_polled_device'
    a_key = 'pattoo_agent_id'

    # Get the device and agent_id that created the datapoints
    metadata = record['pattoo_metadata']
    for item in metadata:
        device = item.get(d_key, None)
        if bool(device) is True:
            break
    for item in metadata:
        agent_id = item.get(a_key, None)
        if bool(agent_id) is True:
            break

    # Create checksum value
    result = data.hashstring('''{}{}{}{}\
'''.format(
        identifier, agent_id, device, record['pattoo_checksum']), sha=512)
    return result
